@@ -3,17 +3,17 @@ import inspect
 import json
 import mimetypes
 import re
-from typing import Any, Callable, Dict, List, Optional, Type
+from typing import Any, Callable, Dict, List, Optional, Type, TypeVar
 
 import pandas as pd
 import streamlit as st
-from pydantic import BaseModel
+from pydantic import BaseModel, ValidationError, parse_obj_as
 from pydantic.json import pydantic_encoder
 
 from streamlit_pydantic import schema_utils
 
 
-def name_to_title(name: str) -> str:
+def _name_to_title(name: str) -> str:
     """Converts a camelCase or snake_case name to title case."""
     # If camelCase -> convert to snake case
     name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
@@ -22,7 +22,7 @@ def name_to_title(name: str) -> str:
     return name.replace("_", " ").strip().title()
 
 
-def function_has_named_arg(func: Callable, parameter: str) -> bool:
+def _function_has_named_arg(func: Callable, parameter: str) -> bool:
     try:
         sig = inspect.signature(func)
         for param in sig.parameters.values():
@@ -33,41 +33,48 @@ def function_has_named_arg(func: Callable, parameter: str) -> bool:
     return False
 
 
-def has_output_ui_renderer(data_item: BaseModel) -> bool:
+def _has_output_ui_renderer(data_item: BaseModel) -> bool:
     return hasattr(data_item, "render_output_ui")
 
 
-def has_input_ui_renderer(input_class: Type[BaseModel]) -> bool:
+def _has_input_ui_renderer(input_class: Type[BaseModel]) -> bool:
     return hasattr(input_class, "render_input_ui")
 
 
-def is_compatible_audio(mime_type: str) -> bool:
+def _is_compatible_audio(mime_type: str) -> bool:
     return mime_type in ["audio/mpeg", "audio/ogg", "audio/wav"]
 
 
-def is_compatible_image(mime_type: str) -> bool:
+def _is_compatible_image(mime_type: str) -> bool:
     return mime_type in ["image/png", "image/jpeg"]
 
 
-def is_compatible_video(mime_type: str) -> bool:
+def _is_compatible_video(mime_type: str) -> bool:
     return mime_type in ["video/mp4"]
 
 
 class InputUI:
+    """Input UI renderer.
+
+    lazydocs: ignore
+    """
+
     def __init__(
         self,
-        streamlit_container: Any,
+        key: str,
         input_class: Type[BaseModel],
-        session_input_key: str = "input_data",
+        streamlit_container: st = st,
         use_sidebar: bool = False,
     ):
+        self._key = key
+
         self._session_state = st.session_state
 
         # Initialize Sessions State
         if "run_id" not in st.session_state:
             self._session_state.run_id = 0
 
-        self._session_input_key = session_input_key
+        self._session_input_key = self._key + "-data"
         if self._session_input_key not in st.session_state:
             self._session_state[self._session_input_key] = {}
 
@@ -84,14 +91,14 @@ class InputUI:
 
         # TODO: check if state has input data
 
-    def render_ui(self) -> None:
-        if has_input_ui_renderer(self._input_class):
+    def render_ui(self) -> Dict:
+        if _has_input_ui_renderer(self._input_class):
             # The input model has a rendering function
             # The rendering also returns the current state of input data
             self._session_state[self._session_input_key] = self._input_class.render_input_ui(  # type: ignore
                 self._streamlit_container, self._session_state[self._session_input_key]
             ).dict()
-            return
+            return self._session_state[self._session_input_key]
 
         required_properties = self._input_class.schema(by_alias=True).get(
             "required", []
@@ -106,7 +113,7 @@ class InputUI:
 
             if not property.get("title"):
                 # Set property key as fallback title
-                property["title"] = name_to_title(property_key)
+                property["title"] = _name_to_title(property_key)
 
             if property_key in required_properties:
                 streamlit_app = self._streamlit_container
@@ -119,10 +126,12 @@ class InputUI:
             except Exception:
                 pass
 
+        return self._session_state[self._session_input_key]
+
     def _get_default_streamlit_input_kwargs(self, key: str, property: Dict) -> Dict:
         streamlit_kwargs = {
             "label": property.get("title"),
-            "key": str(self._session_state.run_id) + "-" + key,
+            "key": str(self._session_state.run_id) + "-" + str(self._key) + "-" + key,
         }
 
         if property.get("description"):
@@ -236,13 +245,13 @@ class InputUI:
 
         bytes = uploaded_file.getvalue()
         if property.get("mime_type"):
-            if is_compatible_audio(property["mime_type"]):
+            if _is_compatible_audio(property["mime_type"]):
                 # Show audio
                 streamlit_app.audio(bytes, format=property.get("mime_type"))
-            if is_compatible_image(property["mime_type"]):
+            if _is_compatible_image(property["mime_type"]):
                 # Show image
                 streamlit_app.image(bytes)
-            if is_compatible_video(property["mime_type"]):
+            if _is_compatible_video(property["mime_type"]):
                 # Show video
                 streamlit_app.video(bytes, format=property.get("mime_type"))
         return bytes
@@ -464,7 +473,7 @@ class InputUI:
             property = properties[property_key]
             if not property.get("title"):
                 # Set property key as fallback title
-                property["title"] = name_to_title(property_key)
+                property["title"] = _name_to_title(property_key)
             # construct full key based on key parts -> required later to get the value
             full_key = key + "." + property_key
             object_inputs[property_key] = self._render_property(
@@ -622,6 +631,11 @@ class InputUI:
 
 
 class OutputUI:
+    """Output UI renderer.
+
+    lazydocs: ignore
+    """
+
     def __init__(self, output_data: Any, input_data: Optional[Any] = None):
         self._output_data = output_data
         self._input_data = input_data
@@ -668,15 +682,15 @@ class OutputUI:
                 mime_type = property_schema["mime_type"]
                 file_extension = mimetypes.guess_extension(mime_type) or ""
 
-                if is_compatible_audio(mime_type):
+                if _is_compatible_audio(mime_type):
                     streamlit.audio(value.as_bytes(), format=mime_type)
                     return
 
-                if is_compatible_image(mime_type):
+                if _is_compatible_image(mime_type):
                     streamlit.image(value.as_bytes())
                     return
 
-                if is_compatible_video(mime_type):
+                if _is_compatible_video(mime_type):
                     streamlit.video(value.as_bytes(), format=mime_type)
                     return
 
@@ -703,8 +717,8 @@ class OutputUI:
 
     def _render_single_output(self, streamlit: st, output_data: BaseModel) -> None:
         try:
-            if has_output_ui_renderer(output_data):
-                if function_has_named_arg(output_data.render_output_ui, "input"):  # type: ignore
+            if _has_output_ui_renderer(output_data):
+                if _function_has_named_arg(output_data.render_output_ui, "input"):  # type: ignore
                     # render method also requests the input data
                     output_data.render_output_ui(streamlit, input=self._input_data)  # type: ignore
                 else:
@@ -731,7 +745,7 @@ class OutputUI:
 
                 output_property_value = output_data.__dict__[property_key]
 
-                if has_output_ui_renderer(output_property_value):
+                if _has_output_ui_renderer(output_property_value):
                     output_property_value.render_output_ui(streamlit)  # type: ignore
                     continue
 
@@ -794,7 +808,7 @@ class OutputUI:
         try:
             data_items: List = []
             for data_item in output_data:
-                if has_output_ui_renderer(data_item):
+                if _has_output_ui_renderer(data_item):
                     # Render using the render function
                     data_item.render_output_ui(streamlit)  # type: ignore
                     continue
@@ -808,23 +822,26 @@ class OutputUI:
 
 
 def pydantic_input(
+    key: str,
     input_class: Type[BaseModel],
-    session_input_key: str = "input_data",
     use_sidebar: bool = False,
-) -> None:
-    """Shows input UI elements for a selected Pydantic class.
+) -> Dict:
+    """Auto-generates input UI elements for a selected Pydantic class.
 
     Args:
+        key (str): A string that identifies the form. Each form must have its own key.
         input_class (Type[BaseModel]): The input class (based on Pydantic BaseModel).
-        session_input_key (str, optional): The key used to store the input data inside the session. Defaults to `input_data`.
         use_sidebar (bool, optional): If `True`, optional input elements will be rendered on the sidebar.
+
+    Returns:
+        Dict: A dictionary with the current state of the input data.
     """
 
-    InputUI(st, input_class, session_input_key=session_input_key).render_ui()
+    return InputUI(key, input_class, use_sidebar=use_sidebar).render_ui()
 
 
 def pydantic_output(output_data: Any) -> None:
-    """Renders the output data based on an instance of a Pydantic model.
+    """Auto-generates output UI elements for all properties of a (Pydantic-based) model instance.
 
     Args:
         output_data (Any): The output data.
@@ -833,4 +850,46 @@ def pydantic_output(output_data: Any) -> None:
     OutputUI(output_data).render_ui()
 
 
-# def pydantic_function(function: )
+# Define generic type to allow autocompletion for the model fields
+T = TypeVar("T", bound=BaseModel)
+
+
+def pydantic_form(
+    key: str,
+    input_class: Type[T],
+    submit_label: str = "Submit",
+    clear_on_submit: bool = False,
+) -> Optional[T]:
+    """Auto-generates a Streamlit form based on the given (Pydantic-based) input class.
+
+    Args:
+        key (str): A string that identifies the form. Each form must have its own key.
+        input_class (Type[BaseModel]): The input class (based on Pydantic BaseModel).
+        submit_label (str): A short label explaining to the user what this button is for. Defaults to “Submit”.
+        clear_on_submit (bool): If True, all widgets inside the form will be reset to their default values after the user presses the Submit button. Defaults to False.
+
+    Returns:
+        Optional[BaseModel]: An instance of the given input class,
+            if the submit button is used and the input data passes the Pydantic validation.
+    """
+
+    with st.form(key=key, clear_on_submit=clear_on_submit):
+        input_state = pydantic_input(key, input_class, use_sidebar=False)
+        submit_button = st.form_submit_button(label=submit_label)
+
+        if submit_button:
+            try:
+                return parse_obj_as(input_class, input_state)
+            except ValidationError as ex:
+                error_text = "**Whoops! There were some problems with your input:**"
+                for error in ex.errors():
+                    if "loc" in error and "msg" in error:
+                        location = ".".join(error["loc"]).replace("__root__.", "")
+                        error_msg = "**" + location + ":** " + error["msg"]
+                        error_text += "\n\n" + error_msg
+                    else:
+                        # Fallback
+                        error_text += "\n\n" + str(error)
+                st.error(error_text)
+                return None
+    return None
